# Should be DB, but week keep it like this to avoid migrating the registry
DATABASE_REGISTRY_DB = "datashare-project-registry"

DATABASE_RUNS_MIGRATION = "_RUNS"
DATABASE_NAME = "name"
# Should be DB, but week keep it like this to avoid migrating the registry
DATABASE_NODE = "_Project"

MIGRATION_COMPLETED = "completed"
MIGRATION_LABEL = "label"
MIGRATION_NODE = "_Migration"
# Should be DB, but week keep it like this to avoid migrating the registry
MIGRATION_DB = "project"
MIGRATION_STARTED = "started"
MIGRATION_STATUS = "status"
MIGRATION_VERSION = "version"
